import logging
from pathlib import Path

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

logger = logging.getLogger(__name__)

configuration = {}


def get_configpath():
    configpaths = [
        Path.cwd() / "configs",
        Path.home() / ".config" / "vm-compact" / "configs",
        Path.home() / "Documents" / "Voicemeeter" / "configs",
    ]
    for configpath in configpaths:
        if configpath.exists():
            return configpath


if configpath := get_configpath():
    filepaths = list(configpath.glob("*.toml"))
    if any(f.stem in ("app", "vban") for f in filepaths):
        configs = {}
        for filepath in filepaths:
            filename = filepath.with_suffix("").stem
            if filename in ("app", "vban"):
                try:
                    with open(filepath, "rb") as f:
                        configs[filename] = tomllib.load(f)
                    logger.info(f"configuration: {filename} loaded into memory")
                except tomllib.TOMLDecodeError:
                    logger.error(f"Invalid TOML config: configs/{filename.stem}")
        configuration |= configs

_defaults = {
    "configs": {
        "config": None,
    },
    "theme": {
        "enabled": True,
        "mode": "light",
    },
    "extends": {
        "extended": True,
        "extends_horizontal": True,
    },
    "channel": {
        "width": 80,
        "height": 130,
        "xpadding": 3,
    },
    "mwscroll_step": {
        "size": 3,
    },
    "submixes": {
        "default": 0,
    },
    "navigation": {"show": True},
}


if "app" in configuration:
    for key in _defaults:
        if key in configuration["app"]:
            configuration["app"][key] = _defaults[key] | configuration["app"][key]
        else:
            configuration["app"][key] = _defaults[key]
else:
    configuration["app"] = _defaults


def get_configuration(key):
    if key in configuration:
        return configuration[key]


def loader(kind_id, target):
    configs = {"reset": target.configs["reset"]}
    if configpath := get_configpath():
        userconfigpath = configpath / kind_id
        if userconfigpath.exists():
            filepaths = list(userconfigpath.glob("*.toml"))
            for filepath in filepaths:
                identifier = filepath.with_suffix("").stem
                try:
                    with open(filepath, "rb") as f:
                        configs[identifier] = tomllib.load(f)
                    logger.info(f"loader: {identifier} loaded into memory")
                except tomllib.TOMLDecodeError:
                    logger.error(f"Invalid TOML config: configs/{filename.stem}")

    target.configs = configs
    return target.configs



import logging
import tkinter as tk
from functools import cached_property
from pathlib import Path
from tkinter import messagebox, ttk
from typing import NamedTuple

import voicemeeterlib

from .builders import MainFrameBuilder
from .configurations import loader
from .data import _base_values, _configuration, _kinds_all, get_configuration
from .errors import VMCompactError
from .menu import Menus
from .subject import Subject

logger = logging.getLogger(__name__)


class App(tk.Tk):
    """App mainframe"""

    @classmethod
    def make(cls, kind: NamedTuple):
        """
        Factory function for App.

        Returns an App class of a kind.
        """

        APP_cls = type(
            f"Voicemeeter{kind}.Compact",
            (cls,),
            {
                "kind": kind,
            },
        )
        return APP_cls

    def __init__(self, vmr):
        super().__init__()
        self.logger = logger.getChild(self.__class__.__name__)
        self._vmr = vmr
        self._vmr.event.add(["pdirty", "ldirty"])
        self.subject = Subject()
        self.start_updates()
        self._vmr.init_thread()
        icon_path = Path(__file__).parent.resolve() / "img" / "cat.ico"
        if icon_path.is_file():
            self.iconbitmap(str(icon_path))
        self.minsize(275, False)
        self._configs = None
        self.protocol("WM_DELETE_WINDOW", self.on_close_window)
        self.menu = self["menu"] = Menus(self, vmr)
        self.styletable = ttk.Style()
        if _configuration.config:
            vmr.apply_config(_configuration.config)

        self.build_app()

        self.drag_id = ""
        self.bind("<Configure>", self.dragging)

        self.after(1, self.healthcheck_step)

    def __str__(self):
        return f"{type(self).__name__}App"

    @property
    def target(self):
        """returns the current interface"""

        return self._vban if _base_values.vban_connected else self._vmr

    @property
    def configframes(self):
        """returns the current configframes"""

        return (
            frame
            for frame in self.winfo_children()
            if isinstance(frame, ttk.Frame)
            and "!stripconfig" in str(frame)
            or "!busconfig" in str(frame)
        )

    def build_app(self, kind=None, vban=None):
        """builds the app frames according to a kind"""
        self._vban = vban
        if kind:
            self.kind = kind

        # register event callbacks
        self.target.subject.add([self.on_pdirty, self.on_ldirty])

        self.bus_frame = None
        self.submix_frame = None
        self.builder = MainFrameBuilder(self)
        self.builder.setup()
        self.builder.create_channelframe("strip")
        self.builder.create_separator()
        self.builder.create_navframe()
        if _configuration.extended:
            self.nav_frame.extend.set(True)
            self.nav_frame.extend_frame()
        if self.kind.name == "potato":
            self.builder.create_banner()

    def on_pdirty(self):
        if _base_values.run_update:
            self.after(1, self.subject.notify, "pdirty")

    def on_ldirty(self):
        if not _base_values.dragging:
            self.after(1, self.subject.notify, "ldirty")

    def _destroy_top_level_frames(self):
        """
        Clear observables.

        Deregister app as observer.

        Destroy all top level frames.
        """
        self.target.subject.remove([self.on_pdirty, self.on_ldirty])
        self.subject.clear()
        [
            frame.destroy()
            for frame in self.winfo_children()
            if isinstance(frame, ttk.Frame)
        ]

    def dragging(self, event, *args):
        if event.widget is self:
            if self.drag_id == "":
                _base_values.dragging = True
            else:
                self.after_cancel(self.drag_id)
            self.drag_id = self.after(100, self.stop_drag)

    def stop_drag(self):
        self.drag_id = ""
        _base_values.dragging = False

    @cached_property
    def userconfigs(self):
        self._configs = loader(self.kind.name, self.target)
        return self._configs

    def start_updates(self):
        def init():
            self.logger.debug("updates started")
            _base_values.run_update = True

        if self._vmr.gui.launched_by_api:
            self.subject.notify("pdirty")
            self.after(12000, init)
        else:
            init()

    def healthcheck_step(self):
        if not _base_values.vban_connected:
            try:
                self._vmr.version
            except voicemeeterlib.error.CAPIError:
                resp = messagebox.askyesno(message="Restart Voicemeeter GUI?")
                if resp:
                    self.logger.debug(
                        "healthcheck failed, rebuilding the app after GUI restart."
                    )
                    self._vmr.end_thread()
                    self._vmr.run_voicemeeter(self._vmr.kind.name)
                    _base_values.run_update = False
                    self._vmr.init_thread()
                    self.after(8000, self.start_updates)
                    self._destroy_top_level_frames()
                    self.build_app(self._vmr.kind)
                    vban_config = get_configuration("vban")
                    for i, _ in enumerate(vban_config):
                        target = getattr(self.menu, f"menu_vban_{i+1}")
                        target.entryconfig(0, state="normal")
                        target.entryconfig(1, state="disabled")
                    [
                        self.menu.menu_vban.entryconfig(j, state="normal")
                        for j, _ in enumerate(self.menu.menu_vban.winfo_children())
                    ]
                else:
                    self.destroy()
        self.after(250, self.healthcheck_step)

    def on_close_window(self):
        if _base_values.vban_connected:
            self._vban.logout()
        self.destroy()


_apps = {kind.name: App.make(kind) for kind in _kinds_all}


def connect(kind_id: str, vmr) -> App:
    """return App of the kind requested"""

    try:
        VMMIN_cls = _apps[kind_id]
    except KeyError:
        raise VMCompactError(f"Invalid kind: {kind_id}")
    return VMMIN_cls(vmr)



import logging
import tkinter as tk
import webbrowser
from functools import partial
from tkinter import messagebox

import vban_cmd
from vban_cmd.error import VBANCMDConnectionError

import sv_ttk

from .data import _base_values, _configuration, get_configuration, kind_get

logger = logging.getLogger(__name__)


class Menus(tk.Menu):
    def __init__(self, parent, vmr):
        super().__init__()
        self.parent = parent
        self.vmr = vmr
        self.logger = logger.getChild(self.__class__.__name__)
        self.vban_config = get_configuration("vban")
        self.app_config = get_configuration("app")
        self._is_topmost = tk.BooleanVar()
        self._lock = tk.BooleanVar()
        self._unlock = tk.BooleanVar()
        self._navigation_show = tk.BooleanVar(value=_configuration.navigation_show)
        self._navigation_hide = tk.BooleanVar(value=not _configuration.navigation_show)
        self._selected_bus = list(tk.BooleanVar() for _ in range(8))

        # voicemeeter menu
        self.menu_voicemeeter = tk.Menu(self, tearoff=0)
        self.add_cascade(menu=self.menu_voicemeeter, label="Voicemeeter")
        self.menu_voicemeeter.add_checkbutton(
            label="Always On Top",
            onvalue=1,
            offvalue=0,
            variable=self._is_topmost,
            command=self.always_on_top,
        )
        self.menu_voicemeeter.add_separator()
        self.menu_voicemeeter.add_command(
            label="Show",
            underline=0,
            command=partial(self.action_invoke_voicemeeter, "show"),
        )
        self.menu_voicemeeter.add_command(
            label="Hide",
            underline=0,
            command=partial(self.action_invoke_voicemeeter, "hide"),
        )
        self.menu_voicemeeter.add_command(
            label="Restart",
            underline=0,
            command=partial(self.action_invoke_voicemeeter, "restart"),
        )
        self.menu_voicemeeter.add_command(
            label="Shutdown",
            underline=0,
            command=partial(self.action_invoke_voicemeeter, "shutdown"),
        )
        self.menu_voicemeeter.add_separator()
        self.menu_lock = tk.Menu(self.menu_voicemeeter, tearoff=0)
        self.menu_voicemeeter.add_cascade(
            menu=self.menu_lock, label="GUI Lock", underline=0
        )
        self.menu_lock.add_checkbutton(
            label="Lock",
            onvalue=1,
            offvalue=0,
            variable=self._lock,
            command=partial(self.action_set_voicemeeter, "lock"),
        )
        self.menu_lock.add_checkbutton(
            label="Unlock",
            onvalue=1,
            offvalue=0,
            variable=self._unlock,
            command=partial(self.action_set_voicemeeter, "lock", False),
        )

        # configs menu
        self.menu_configs = tk.Menu(self, tearoff=0)
        self.add_cascade(menu=self.menu_configs, label="Configs")
        self.menu_configs_load = tk.Menu(self.menu_configs, tearoff=0)
        self.menu_configs.add_cascade(menu=self.menu_configs_load, label="Load config")
        self.config_defaults = {"reset"}
        if len(self.parent.userconfigs) > len(self.config_defaults) and all(
            key in self.parent.userconfigs for key in self.config_defaults
        ):
            [
                self.menu_configs_load.add_command(
                    label=profile, command=partial(self.load_profile, profile)
                )
                for profile in self.parent.userconfigs.keys()
                if profile not in self.config_defaults
            ]
        else:
            self.menu_configs.entryconfig(0, state="disabled")
        self.menu_configs.add_command(
            label="Reset to defaults", command=self.load_defaults
        )

        # layout menu
        self.menu_layout = tk.Menu(self, tearoff=0)
        self.add_cascade(menu=self.menu_layout, label="Layout")
        # layout/submixes
        # here we build menu regardless of kind but disable if not potato
        buses = tuple(f"A{i+1}" for i in range(5)) + tuple(f"B{i+1}" for i in range(3))
        self.menu_submixes = tk.Menu(self.menu_layout, tearoff=0)
        self.menu_layout.add_cascade(menu=self.menu_submixes, label="Submixes")
        [
            self.menu_submixes.add_checkbutton(
                label=f"Bus {buses[i]}",
                underline=0,
                onvalue=1,
                offvalue=0,
                variable=self._selected_bus[i],
                command=partial(self.set_submix, i),
            )
            for i in range(8)
        ]
        self._selected_bus[_configuration.submixes].set(True)
        if self.parent.kind.name != "potato":
            self.menu_layout.entryconfig(0, state="disabled")
        # layout/extends
        self.menu_extends = tk.Menu(self.menu_layout, tearoff=0)
        self.menu_layout.add_cascade(
            menu=self.menu_extends, label="Extends", underline=0
        )
        self.menu_extends.add_command(
            label="horizontal",
            underline=0,
            command=partial(self.switch_orientation, extends_horizontal=True),
        )
        self.menu_extends.add_command(
            label="vertical",
            underline=0,
            command=partial(self.switch_orientation, extends_horizontal=False),
        )
        self.menu_extends.entryconfig(
            0 if _configuration.extends_horizontal else 1, state="disabled"
        )
        # layout/themes
        self.menu_themes = tk.Menu(self.menu_layout, tearoff=0)
        self.menu_layout.add_cascade(menu=self.menu_themes, label="Themes")
        self.menu_themes.add_command(
            label="light", command=partial(self.load_theme, "light")
        )
        self.menu_themes.add_command(
            label="dark", command=partial(self.load_theme, "dark")
        )
        self.menu_themes.entryconfig(
            0 if self.app_config["theme"]["mode"] == "light" else 1,
            state="disabled",
        )
        if not _configuration.themes_enabled:
            self.menu_layout.entryconfig(2, state="disabled")
        # layout/navigation
        self.menu_navigation = tk.Menu(self.menu_layout, tearoff=0)
        self.menu_layout.add_cascade(menu=self.menu_navigation, label="Navigation")
        self.menu_navigation.add_checkbutton(
            label="show",
            onvalue=1,
            offvalue=0,
            variable=self._navigation_show,
            command=partial(self.toggle_navigation, "show"),
        )
        self.menu_navigation.add_checkbutton(
            label="hide",
            onvalue=1,
            offvalue=0,
            variable=self._navigation_hide,
            command=partial(self.toggle_navigation, "hide"),
        )

        # vban connect menu
        self.menu_vban = tk.Menu(self, tearoff=0)
        self.add_cascade(menu=self.menu_vban, label="VBAN")
        if self.vban_config:
            for i, _ in enumerate(self.vban_config):
                setattr(self, f"menu_vban_{i+1}", tk.Menu(self.menu_vban, tearoff=0))
                target_menu = getattr(self, f"menu_vban_{i+1}")
                self.menu_vban.add_cascade(
                    menu=target_menu,
                    label=f"{self.vban_config[f'connection-{i+1}']['streamname']}",
                    underline=0,
                )
                target_menu.add_command(
                    label="Connect", command=partial(self.vban_connect, i)
                )
                target_menu.add_command(
                    label="Disconnect", command=partial(self.vban_disconnect, i)
                )
                target_menu.entryconfig(1, state="disabled")
        else:
            self.entryconfig(4, state="disabled")

        # Help menu
        self.menu_help = tk.Menu(self, tearoff=0)
        self.add_cascade(menu=self.menu_help, label="Help")
        self.menu_help.add_command(
            label="Voicemeeter Site",
            command=self.documentation,
        )
        self.menu_help.add_command(
            label="Source Code",
            command=self.github,
        )
        self.menu_help.add_command(
            label="App Creator",
            command=self.onyxandiris,
        )

    @property
    def target(self):
        """use the correct interface"""
        return self.parent.target

    def enable_vban_menus(self):
        [
            self.menu_vban.entryconfig(j, state="normal")
            for j, _ in enumerate(self.menu_vban.winfo_children())
        ]

    def action_invoke_voicemeeter(self, cmd):
        if fn := getattr(self.target.command, cmd):
            fn()
        if cmd == "shutdown":
            self.parent.on_close_window()

    def action_set_voicemeeter(self, cmd, val=True):
        if cmd == "lock":
            self._lock.set(val)
            self._unlock.set(not self._lock.get())
        setattr(self.target.command, cmd, val)

    def load_custom_profile(self, profile):
        self.logger.info(f"loading user profile {profile}")
        self.target.apply(profile)
        if not _base_values.run_update:
            self.parent.subject.notify("pdirty")

    def load_profile(self, profile):
        self.logger.info(f"loading user profile {profile}")
        self.target.apply_config(profile)
        if not _base_values.run_update:
            self.parent.subject.notify("pdirty")

    def load_defaults(self):
        msg = (
            "Are you sure you want to Reset values to defaults?",
            "Physical strips B1, Virtual strips A1",
            "Mono, Solo, Mute, EQ all OFF",
            "Gain sliders for Strip/Bus at 0.0",
        )
        resp = messagebox.askyesno(message="\n".join(msg))
        if resp:
            self.load_profile("reset")

    def always_on_top(self):
        self.parent.attributes("-topmost", self._is_topmost.get())

    def switch_orientation(self, extends_horizontal: bool = True, *args):
        _configuration.extends_horizontal = extends_horizontal
        if extends_horizontal:
            self.menu_extends.entryconfig(0, state="disabled")
            self.menu_extends.entryconfig(1, state="normal")
        else:
            self.menu_extends.entryconfig(1, state="disabled")
            self.menu_extends.entryconfig(0, state="normal")

    def set_submix(self, i):
        if _configuration.submixes != i:
            _configuration.submixes = i
            if self.parent.submix_frame is not None:
                self.parent.submix_frame.teardown()
                self.parent.nav_frame.show_submix()
            for j, var in enumerate(self._selected_bus):
                var.set(i == j)
            self.parent.subject.notify("submix")

    def load_theme(self, theme):
        sv_ttk.set_theme(theme)
        _configuration.theme_mode = theme
        self.menu_themes.entryconfig(
            0,
            state=f"{'disabled' if theme == 'light' else 'normal'}",
        )
        self.menu_themes.entryconfig(
            1,
            state=f"{'disabled' if theme == 'dark' else 'normal'}",
        )
        [
            menu.config(bg=f"{'black' if theme == 'dark' else 'white'}")
            for menu in self.winfo_children()
            if isinstance(menu, tk.Menu)
        ]
        self.menu_lock.config(bg=f"{'black' if theme == 'dark' else 'white'}")
        self.menu_configs_load.config(bg=f"{'black' if theme == 'dark' else 'white'}")
        [
            menu.config(bg=f"{'black' if theme == 'dark' else 'white'}")
            for menu in self.menu_vban.winfo_children()
            if isinstance(menu, tk.Menu)
        ]
        [
            menu.config(bg=f"{'black' if theme == 'dark' else 'white'}")
            for menu in self.menu_layout.winfo_children()
            if isinstance(menu, tk.Menu)
        ]
        self.logger.info(
            f"Finished loading theme Sunvalley {sv_ttk.get_theme().capitalize()} theme"
        )

    def menu_teardown(self, i):
        # remove config load menus
        if len(self.parent.userconfigs) > len(self.config_defaults):
            for profile in self.parent.userconfigs:
                if profile not in self.config_defaults:
                    try:
                        self.menu_configs_load.delete(profile)
                    except tk._tkinter.tclError as e:
                        self.logger.warning(f"{type(e).__name__}: {e}")

        [
            self.menu_vban.entryconfig(j, state="disabled")
            for j, _ in enumerate(self.menu_vban.winfo_children())
            if j != i
        ]

    def menu_setup(self):
        if len(self.parent.userconfigs) > len(self.config_defaults):
            for profile in self.parent.userconfigs:
                if profile not in self.config_defaults:
                    self.menu_configs_load.add_command(
                        label=profile, command=partial(self.load_profile, profile)
                    )
                    self.menu_configs.entryconfig(0, state="normal")
        else:
            self.menu_configs.entryconfig(0, state="disabled")

    def toggle_navigation(self, cmd=None):
        if cmd == "show":
            self.logger.debug("show navframe")
            self.parent.nav_frame.grid()
            self._navigation_show.set(True)
            self._navigation_hide.set(not self._navigation_show.get())
        else:
            self.logger.debug("hide navframe")
            self.parent.nav_frame.grid_remove()
            self._navigation_hide.set(True)
            self._navigation_show.set(not self._navigation_hide.get())

    def vban_connect(self, i):
        opts = {}
        opts |= self.vban_config[f"connection-{i+1}"]
        kind_id = opts.pop("kind")
        self.vban = vban_cmd.api(kind_id, **opts)
        # login to vban interface
        try:
            self.logger.info(f"Attempting vban connection to {opts.get('ip')}")
            self.vban.login()
        except VBANCMDConnectionError as e:
            self.vban.logout()
            msg = (
                f"Timeout attempting to establish connection to {opts.get('ip')}",
                f"Please check your connection settings",
            )
            messagebox.showerror("Connection Error", "\n".join(msg))
            msg = (str(e), f"resuming local connection")
            self.logger.error(", ".join(msg))
            self.after(1, self.enable_vban_menus)
            return
        self.menu_teardown(i)
        self.vban.event.add(["pdirty", "ldirty"])
        # destroy the current App frames
        self.parent._destroy_top_level_frames()
        _base_values.vban_connected = True
        self.vmr.end_thread()

        # build new app frames according to a kind
        kind = kind_get(kind_id)
        self.parent.build_app(kind, self.vban)
        target_menu = getattr(self, f"menu_vban_{i+1}")
        target_menu.entryconfig(0, state="disabled")
        target_menu.entryconfig(1, state="normal")
        self.menu_layout.entryconfig(
            0, state=f"{'normal' if kind.name == 'potato' else 'disabled'}"
        )
        # ensure the configs are reloaded into memory
        if "config" in self.parent.target.__dict__:
            del self.parent.target.__dict__["config"]
        if "userconfigs" in self.parent.__dict__:
            del self.parent.__dict__["userconfigs"]
        self.menu_setup()

    def vban_disconnect(self, i):
        self.menu_teardown(i)

        # destroy the current App frames
        self.parent._destroy_top_level_frames()
        _base_values.vban_connected = False
        # logout of vban interface
        self.vmr.init_thread()
        self.vban.logout()
        # build new app frames according to a kind
        kind = kind_get(self.vmr.type)
        self.parent.build_app(kind)
        target_menu = getattr(self, f"menu_vban_{i+1}")
        target_menu.entryconfig(0, state="normal")
        target_menu.entryconfig(1, state="disabled")
        self.menu_layout.entryconfig(
            0, state=f"{'normal' if kind.name == 'potato' else 'disabled'}"
        )
        # ensure the configs are reloaded into memory
        if "config" in self.parent.target.__dict__:
            del self.parent.target.__dict__["config"]
        if "userconfigs" in self.parent.__dict__:
            del self.parent.__dict__["userconfigs"]
        self.menu_setup()

        self.after(15000, self.enable_vban_menus)

    def documentation(self):
        webbrowser.open_new(r"https://voicemeeter.com/")

    def github(self):
        webbrowser.open_new(r"https://github.com/onyx-and-iris/voicemeeter-compact")

    def onyxandiris(self):
        webbrowser.open_new(r"https://onyxandiris.online")



import logging
from tkinter import ttk

from . import builders
from .data import _base_values, _configuration

logger = logging.getLogger(__name__)


class Config(ttk.Frame):
    def __init__(self, parent, index, _id):
        super().__init__(parent)
        self.parent = parent
        self.index = index
        self.id = _id
        self.logger = logger.getChild(self.__class__.__name__)
        self.styletable = parent.styletable
        self.phys_in, self.virt_in = parent.kind.ins
        self.phys_out, self.virt_out = parent.kind.outs

        self.parent.subject.add(self)

    @property
    def identifier(self):
        return self.id

    @property
    def target(self):
        """returns the current interface"""

        return self.parent.target

    def getter(self, param):
        param = param.split(".")
        try:
            if len(param) == 2:
                target = getattr(self.target, param[0])
                return getattr(target, param[1])
            else:
                return getattr(self.target, param[0])
        except AttributeError as e:
            self.logger.error(f"{type(e).__name__}: {e}")

    def setter(self, param, value):
        param = param.split(".")
        try:
            if len(param) == 2:
                target = getattr(self.target, param[0])
                setattr(target, param[1], value)
            else:
                setattr(self.target, param[0], value)
        except AttributeError as e:
            self.logger(f"{type(e).__name__}: {e}")

    def scale_press(self, *args):
        self.after(1, self.remove_events)

    def remove_events(self):
        self.parent.target.event.remove("pdirty")
        self.parent.target.event.remove("ldirty")

    def scale_release(self, *args):
        _base_values.run_update = False
        self.after(1, self.add_events)

    def add_events(self):
        self.parent.target.event.add("pdirty")
        self.parent.target.event.add("ldirty")
        self.after(350, self.resume_updates)

    def pause_updates(self, func, *args):
        """function wrapper, adds a 50ms delay on updates"""
        _base_values.run_update = False

        func(*args)

        self.after(50, self.resume_updates)

    def resume_updates(self):
        _base_values.run_update = True

    def scale_enter(self, param, *args):
        val = self.slider_vars[self.slider_params.index(param)].get()
        self.parent.nav_frame.info_text.set(round(val, 1))

    def scale_leave(self, *args):
        self.parent.nav_frame.info_text.set("")

    def scale_callback(self, param, *args):
        """callback function for scale widget"""

        val = self.slider_vars[self.slider_params.index(param)].get()
        self.setter(param, round(val, 1))
        self.parent.nav_frame.info_text.set(round(val, 1))

    def reset_scale(self, param, val, *args):
        self.setter(param, val)
        self.slider_vars[self.slider_params.index(param)].set(val)

    def toggle_p(self, param):
        val = self.param_vars[self.params.index(param)].get()
        self.setter(param, val)
        if not _configuration.themes_enabled:
            self.styletable.configure(
                f"{param}.TButton", background=f'{"green" if val else "white"}'
            )

    def on_update(self, subject):
        """update parameters"""
        if subject == "pdirty":
            self.sync()


class StripConfig(Config):
    def __init__(self, parent, index, _id):
        super().__init__(parent, index, _id)
        self.grid(column=0, row=1, columnspan=4, padx=(2,))
        self.builder = builders.StripConfigFrameBuilder(self)
        self.builder.setup()
        self.make_row_0()
        self.make_row_1()
        self.make_row_2()
        self.builder.grid_configure()

        self.parent.target.clear_dirty()
        self.sync()

    @property
    def target(self):
        """returns the strip class for this configframe in the current interface"""

        _target = super(StripConfig, self).target
        return getattr(_target, self.identifier)[self.index]

    def make_row_0(self):
        if self.index < self.phys_in:
            if self.parent.kind.name == "basic":
                self.builder.create_audibility_slider()
            else:
                self.builder.create_comp_slider()
                self.builder.create_gate_slider()
                self.builder.create_limit_slider()

    def make_row_1(self):
        self.builder.create_a_buttons()
        self.builder.create_b_buttons()

    def make_row_2(self):
        self.builder.create_param_buttons()

    def toggle_a(self, param):
        val = self.phys_out_params_vars[self.phys_out_params.index(param)].get()
        self.setter(param, val)
        if not _configuration.themes_enabled:
            self.styletable.configure(
                f"{param}.TButton", background=f'{"green" if val else "white"}'
            )

    def toggle_b(self, param):
        val = self.virt_out_params_vars[self.virt_out_params.index(param)].get()
        self.setter(param, val)
        if not _configuration.themes_enabled:
            self.styletable.configure(
                f"{param}.TButton", background=f'{"green" if val else "white"}'
            )

    def teardown(self):
        self.builder.teardown()

    def sync(self):
        [
            self.phys_out_params_vars[i].set(self.getter(param))
            for i, param in enumerate(self.phys_out_params)
        ]
        [
            self.virt_out_params_vars[i].set(self.getter(param))
            for i, param in enumerate(self.virt_out_params)
        ]
        [
            self.param_vars[i].set(self.getter(param))
            for i, param in enumerate(self.params)
        ]
        if not _base_values.vban_connected:  # slider vars not defined in RT Packet
            [
                self.slider_vars[i].set(self.getter(param))
                for i, param in enumerate(self.slider_params)
                if self.index < self.phys_in
            ]

        if not _configuration.themes_enabled:
            [
                self.styletable.configure(
                    f"{param}.TButton",
                    background=f'{"green" if self.phys_out_params_vars[i].get() else "white"}',
                )
                for i, param in enumerate(self.phys_out_params)
            ]
            [
                self.styletable.configure(
                    f"{param}.TButton",
                    background=f'{"green" if self.virt_out_params_vars[i].get() else "white"}',
                )
                for i, param in enumerate(self.virt_out_params)
            ]
            [
                self.styletable.configure(
                    f"{param}.TButton",
                    background=f'{"green" if self.param_vars[i].get() else "white"}',
                )
                for i, param in enumerate(self.params)
            ]


class BusConfig(Config):
    def __init__(self, parent, index, _id):
        super().__init__(parent, index, _id)
        if _configuration.extends_horizontal:
            self.grid(column=0, row=1, columnspan=4, padx=(2,))
        else:
            self.grid(column=0, row=3, columnspan=4, padx=(2,))
        self.builder = builders.BusConfigFrameBuilder(self)
        self.builder.setup()
        self.make_row_0()
        self.make_row_1()
        self.builder.grid_configure()

        self.parent.target.clear_dirty()
        self.sync()

    @property
    def target(self):
        """returns the bus class for this configframe in the current interface"""

        _target = super(BusConfig, self).target
        return getattr(_target, self.identifier)[self.index]

    def make_row_0(self):
        self.builder.create_bus_mode_button()

    def make_row_1(self):
        self.builder.create_param_buttons()

    def current_bus_mode(self):
        for mode in self.bus_modes:
            if getattr(self.target.mode, mode):
                return mode

    def rotate_bus_modes_right(self, *args):
        current_mode = self.current_bus_mode()
        next = self.bus_modes.index(current_mode) + 1
        if next < len(self.bus_modes):
            setattr(
                self.target.mode,
                self.bus_modes[next],
                True,
            )
            self.bus_mode_label_text.set(self.bus_mode_map[self.bus_modes[next]])
        else:
            self.target.mode.normal = True
            self.bus_mode_label_text.set("Normal")

    def rotate_bus_modes_left(self, *args):
        current_mode = self.current_bus_mode()
        prev = self.bus_modes.index(current_mode) - 1
        if prev < 0:
            self.target.mode.rearonly = True
            self.bus_mode_label_text.set("Rear Only")
        else:
            setattr(
                self.target.mode,
                self.bus_modes[prev],
                True,
            )
            self.bus_mode_label_text.set(self.bus_mode_map[self.bus_modes[prev]])

    def teardown(self):
        self.builder.teardown()

    def sync(self):
        [
            self.param_vars[i].set(self.getter(param))
            for i, param in enumerate(self.params)
        ]
        self.bus_mode_label_text.set(self.bus_mode_map[self.current_bus_mode()])
        if not _configuration.themes_enabled:
            [
                self.styletable.configure(
                    f"{param}.TButton",
                    background=f'{"green" if self.param_vars[i].get() else "white"}',
                )
                for i, param in enumerate(self.params)
            ]



import logging
import tkinter as tk
from tkinter import ttk

from .data import _base_values, _configuration

logger = logging.getLogger(__name__)


class Banner(ttk.Frame):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.parent.subject.add(self)
        self.logger = logger.getChild(self.__class__.__name__)
        self.submix = tk.StringVar(value=self.target.bus[_configuration.submixes].label)

        self.label = ttk.Label(
            self,
            text=f"SUBMIX: {self.submix.get().upper()}",
        )
        self.label.grid(column=0, row=0, sticky=(tk.N, tk.S, tk.W, tk.E))

    @property
    def target(self):
        """returns the current interface"""

        return self.parent.target

    def on_update(self, subject):
        if subject == "submix":
            if not _base_values.dragging:
                self.logger.debug("checking submix for banner")
                self.submix.set(self.target.bus[_configuration.submixes].label)
                self.label["text"] = f"SUBMIX: {self.submix.get().upper()}"



import logging
import tkinter as tk
from tkinter import ttk

from . import builders
from .data import _configuration
from .gainlayer import SubMixFrame

logger = logging.getLogger(__name__)


class Navigation(ttk.Frame):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.logger = logger.getChild(self.__class__.__name__)
        self.grid(row=0, column=3, padx=(0, 2), pady=(5, 5), sticky=(tk.W, tk.E))
        if not _configuration.navigation_show:
            self.grid_remove()
        self.styletable = self.parent.styletable

        self.builder = builders.NavigationFrameBuilder(self)
        self.builder.setup()
        self.builder.create_submix_button()
        self.builder.create_channel_button()
        self.builder.create_extend_button()
        self.builder.create_info_button()
        self.builder.grid_configure()

        self.mainframebuilder = builders.MainFrameBuilder(self.parent)

    def show_submix(self):
        if self.submix.get():
            self.parent.submix_frame = SubMixFrame(self.parent)
            self.logger.info(
                f"Finished building submixframe for submix {_configuration.submixes}"
            )
        else:
            if _configuration.extends_horizontal:
                self.parent.submix_frame.teardown()
                if self.parent.bus_frame:
                    self.parent.bus_frame.grid()
                else:
                    self.parent.columnconfigure(1, weight=0)
            else:
                self.parent.submix_frame.teardown()
                if self.parent.bus_frame:
                    self.parent.bus_frame.grid()
                else:
                    self.parent.rowconfigure(2, weight=0, minsize=0)
            self.logger.info(
                f"Finished tearing down submixframe for submix {_configuration.submixes}"
            )

        if not _configuration.themes_enabled:
            self.styletable.configure(
                f"Submix.TButton",
                background=f'{"purple" if self.submix.get() else "white"}',
            )

    def switch_channel(self):
        if self.channel_text.get() == "STRIP":
            self.mainframebuilder.create_channelframe("bus")
            self.parent.strip_frame.teardown()
        else:
            self.mainframebuilder.create_channelframe("strip")
            self.parent.bus_frame.teardown()

        self.extend_button["state"] = (
            "disabled" if self.channel_text.get() == "STRIP" else "normal"
        )
        [frame.teardown() for frame in self.parent.configframes]
        self.channel_text.set("BUS" if self.channel_text.get() == "STRIP" else "STRIP")

    def extend_frame(self):
        _configuration.extended = self.extend.get()
        if self.extend.get():
            self.channel_button["state"] = "disabled"
            self.mainframebuilder.create_channelframe("bus")
        else:
            [
                frame.teardown()
                for frame in self.parent.configframes
                if "!busconfig" in str(frame)
            ]
            self.parent.bus_frame.teardown()
            self.parent.bus_frame = None
            self.channel_button["state"] = "normal"

            if self.parent.submix_frame:
                self.parent.submix_frame.teardown()
                self.submix.set(False)
                if not _configuration.themes_enabled:
                    self.styletable.configure(
                        f"Submix.TButton",
                        background=f'{"purple" if self.submix.get() else "white"}',
                    )

        self.extend_text.set("REDUCE" if self.extend.get() else "EXTEND")



import logging
import tkinter as tk
from tkinter import ttk

from . import builders
from .data import _base_values, _configuration

logger = logging.getLogger(__name__)


class ChannelLabelFrame(ttk.LabelFrame):
    """Base class for a single channel"""

    def __init__(self, parent, index, id):
        super().__init__(parent)
        self.parent = parent
        self.index = index
        self.id = id
        self.logger = logger.getChild(self.__class__.__name__)
        self.styletable = self.parent.parent.styletable

        self.builder = builders.ChannelLabelFrameBuilder(self, index, id)
        self.builder.setup()
        self.builder.add_progressbar()
        self.builder.add_scale()
        self.builder.add_mute_button()
        self.builder.add_conf_button()
        self.builder.add_gain_label()
        self.sync_params()
        self.sync_labels()
        self.grid_configure()

        self.configbuilder = builders.MainFrameBuilder(self.parent.parent)

    @property
    def identifier(self):
        return self.id

    @property
    def target(self):
        """returns the current interface"""

        return self.parent.target

    def getter(self, param):
        try:
            return getattr(self.target, param)
        except AttributeError as e:
            self.logger(f"{type(e).__name__}: {e}")

    def setter(self, param, value):
        if param in dir(self.target):  # avoid calling getattr (with hasattr)
            setattr(self.target, param, value)

    def scale_callback(self, *args):
        """callback function for scale widget"""

        val = round(self.gain.get(), 1)
        self.setter("gain", val)
        self.gainlabel.set(val)

    def toggle_mute(self, *args):
        self.target.mute = self.mute.get()
        if not _configuration.themes_enabled:
            self.styletable.configure(
                f"{self.identifier}Mute{self.index}.TButton",
                background=f'{"red" if self.mute.get() else "white"}',
            )

    def reset_gain(self, *args):
        self.setter("gain", 0)
        self.gain.set(0)
        self.gainlabel.set(self.gain.get())

    def scale_press(self, *args):
        self.after(1, self.remove_events)

    def remove_events(self):
        self.parent.target.event.remove("pdirty")
        self.parent.target.event.remove("ldirty")

    def scale_release(self, *args):
        _base_values.run_update = False
        self.after(1, self.add_events)

    def add_events(self):
        self.parent.target.event.add("pdirty")
        self.parent.target.event.add("ldirty")
        self.after(500, self.resume_updates)

    def pause_updates(self, func, *args):
        """function wrapper, adds a 50ms delay on updates"""
        _base_values.run_update = False

        func(*args)

        self.after(50, self.resume_updates)

    def resume_updates(self):
        _base_values.run_update = True

    def _on_mousewheel(self, event):
        self.gain.set(
            round(
                self.gain.get()
                + (
                    _configuration.mwscroll_step
                    if event.delta > 0
                    else -_configuration.mwscroll_step
                ),
                1,
            )
        )
        if self.gain.get() > 12:
            self.gain.set(12)
        elif self.gain.get() < -60:
            self.gain.set(-60)
        self.setter("gain", self.gain.get())
        self.gainlabel.set(round(self.gain.get(), 1))

    def open_config(self):
        if self.conf.get():
            self.configbuilder.create_configframe(self.identifier, self.index, self.id)
        else:
            self.parent.parent.config_frame.teardown()
        if not _configuration.themes_enabled:
            self.styletable.configure(
                f"{self.identifier}Conf{self.index}.TButton",
                background=f'{"yellow" if self.conf.get() else "white"}',
            )

    def on_update(self, subject):
        if subject == "ldirty":
            self.upd_levels()
        elif subject == "pdirty":
            self.sync_params()
        elif subject == "labelframe":
            self.after(5, self.sync_labels)

    def sync_params(self):
        """sync parameter states, update button colours"""
        self.gain.set(self.getter("gain"))
        self.gainlabel.set(round(self.gain.get(), 1))
        self.mute.set(self.getter("mute"))
        if not _configuration.themes_enabled:
            self.styletable.configure(
                f"{self.identifier}Mute{self.index}.TButton",
                background=f'{"red" if  self.mute.get() else "white"}',
            )

    def sync_labels(self):
        """sync labelframes according to label text"""
        retval = self.getter("label")
        if self.parent.label_cache[self.id][self.index] != retval:
            self.parent.label_cache[self.id][self.index] = retval
            if len(retval) > 10:
                retval = f"{retval[:8]}.."
            if not retval:
                self.parent.columnconfigure(self.index, minsize=0)
                self.parent.parent.subject.remove(self)
                self.grid_remove()
            else:
                self.parent.parent.subject.add(self)
                self.grid()
            self.configure(text=retval)

    def grid_configure(self):
        self.grid(padx=_configuration.channel_xpadding, sticky=(tk.N, tk.S))
        [
            child.grid_configure(padx=1, pady=1, sticky=(tk.W, tk.E))
            for child in self.winfo_children()
            if isinstance(child, ttk.Checkbutton)
        ]
        [
            child.grid_configure(padx=1, pady=1, sticky=(tk.N, tk.S))
            for child in self.winfo_children()
            if isinstance(child, ttk.Progressbar) or isinstance(child, ttk.Scale)
        ]


class Strip(ChannelLabelFrame):
    """Concrete class representing a single strip"""

    def __init__(self, parent, index, id):
        super().__init__(parent, index, id)
        if index <= parent.phys_in:
            self.level_offset = index * 2
        else:
            self.level_offset = parent.phys_in * 2 + (index - parent.phys_in) * 8

    @property
    def target(self):
        """returns the strip class for this labelframe in the current interface"""

        _target = super(Strip, self).target
        return getattr(_target, self.identifier)[self.index]

    def upd_levels(self):
        """
        Updates level values.
        """
        if self.index < self.parent.parent.kind.num_strip:
            if self.target.levels.is_updated:
                val = max(self.target.levels.prefader)
                self.level.set(
                    (0 if self.mute.get() else 72 + val - 12 + self.gain.get())
                )


class Bus(ChannelLabelFrame):
    """Concrete bus class representing a single bus"""

    def __init__(self, parent, index, id):
        super().__init__(parent, index, id)
        self.level_offset = index * 8

    @property
    def target(self):
        """returns the bus class for this labelframe in the current interface"""

        _target = super(Bus, self).target
        return getattr(_target, self.identifier)[self.index]

    def upd_levels(self):
        if self.index < self.parent.parent.kind.num_bus:
            if self.target.levels.is_updated or self.level.get() != -118:
                val = max(self.target.levels.all)
                self.level.set((0 if self.mute.get() else 72 + val - 12))


class ChannelFrame(ttk.Frame):
    def init(self, parent, id):
        super().__init__(parent)
        self.parent = parent
        self.id = id
        self.phys_in, self.virt_in = parent.kind.ins
        self.phys_out, self.virt_out = parent.kind.outs
        self.label_cache = {
            "strip": [""] * (self.phys_in + self.virt_in),
            "bus": [""] * (self.phys_out + self.virt_out),
        }
        self.parent.subject.add(self)
        self.update_labels()

    @property
    def target(self):
        """returns the current interface"""

        return self.parent.target

    @property
    def identifier(self):
        return self.id

    @property
    def labelframes(self):
        """returns a tuple of current channel labelframe addresses"""

        return tuple(
            frame
            for frame in self.winfo_children()
            if isinstance(frame, ttk.LabelFrame)
        )

    def update_labels(self):
        for labelframe in self.labelframes:
            labelframe.on_update("labelframe")

    def on_update(self, subject):
        if subject == "pdirty":
            self.update_labels()

    def grid_configure(self):
        [
            self.columnconfigure(i, minsize=_configuration.channel_width)
            for i, _ in enumerate(self.labelframes)
        ]
        [self.rowconfigure(0, minsize=100) for i, _ in enumerate(self.labelframes)]

    def teardown(self):
        [self.parent.subject.remove(frame) for frame in self.labelframes]
        self.parent.subject.remove(self)
        self.destroy()
        setattr(self.parent, f"{self.identifier}_frame", None)


def _make_channelframe(parent, identifier):
    """
    Creates a Channel Frame class of type strip or bus
    """

    phys_in, virt_in = parent.kind.ins
    phys_out, virt_out = parent.kind.outs

    def init_labels(self):
        """
        Grids each labelframe, grid_removes any without a label
        """

        for i, labelframe in enumerate(
            getattr(self, "strips" if identifier == "strip" else "buses")
        ):
            labelframe.grid(row=0, column=i)
            label = labelframe.target.label
            if not label:
                self.columnconfigure(i, minsize=0)
                labelframe.grid_remove()
            self.label_cache[identifier][i] = label

    def init_strip(self, *args, **kwargs):
        self.init(parent, identifier)
        self.strips = tuple(
            Strip(self, i, identifier) for i in range(phys_in + virt_in)
        )
        self.grid(row=0, column=0, sticky=(tk.W))
        self.grid_configure()
        init_labels(self)

    def init_bus(self, *args, **kwargs):
        self.init(parent, identifier)
        self.buses = tuple(Bus(self, i, identifier) for i in range(phys_out + virt_out))
        if _configuration.extended:
            if _configuration.extends_horizontal:
                self.grid(row=0, column=2, sticky=(tk.W))
            else:
                self.grid(row=2, column=0, sticky=(tk.W))
        else:
            self.grid(row=0, column=0)
        self.grid_configure()
        init_labels(self)

    if identifier == "strip":
        CHANNELFRAME_cls = type(
            f"ChannelFrame{identifier.capitalize()}",
            (ChannelFrame,),
            {
                "__init__": init_strip,
            },
        )
    else:
        CHANNELFRAME_cls = type(
            f"ChannelFrame{identifier.capitalize()}",
            (ChannelFrame,),
            {
                "__init__": init_bus,
            },
        )
    return CHANNELFRAME_cls(parent)



from dataclasses import dataclass

from voicemeeterlib import kinds

from .configurations import get_configuration

configuration = get_configuration("app")


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


@dataclass
class Configurations(metaclass=SingletonMeta):
    # is the gui extended
    extended: bool = configuration["extends"]["extended"]
    # direction the gui extends
    extends_horizontal: bool = configuration["extends"]["extends_horizontal"]
    # are themes enabled
    themes_enabled: bool = configuration["theme"]["enabled"]
    # light or dark
    theme_mode: str = configuration["theme"]["mode"]
    # size of mousewheel scroll step
    mwscroll_step: int = configuration["mwscroll_step"]["size"]
    # bus assigned as current submix
    submixes: int = configuration["submixes"]["default"]

    # width of a single channel labelframe
    channel_width: int = configuration["channel"]["width"]
    # height of a single channel labelframe
    channel_height: int = configuration["channel"]["height"]
    # xpadding for a single channel labelframe
    channel_xpadding: int = configuration["channel"]["xpadding"]

    # do we grid the navigation frame?
    navigation_show: bool = configuration["navigation"]["show"]

    @property
    def config(self):
        if "configs" in configuration:
            return configuration["configs"]["config"]


@dataclass
class BaseValues(metaclass=SingletonMeta):
    # pause updates after releasing scale
    run_update: bool = False
    # are we dragging main window with mouse 1
    dragging: bool = False
    # a vban connection established
    vban_connected: bool = False


_base_values = BaseValues()
_configuration = Configurations()

_kinds = {kind.name: kind for kind in kinds.kinds_all}

_kinds_all = _kinds.values()


def kind_get(kind_id):
    return _kinds[kind_id]
